#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <SPI.h>
#include <MFRC522.h>
#include <SD.h>
#include <Servo.h>

#define RST_PIN 9 
#define MQ3pin A0 
#define PASSIVE_BUZZER_PIN 2 
#define SERVO_PIN 6 

MFRC522 mfrc522; 
LiquidCrystal_I2C lcd(0x27, 16, 2); 
Servo myservo; 

const int chipSelect = 4; // SD card chip select pin

File dataFile;

void setup() {
  Serial.begin(9600); // Initialize serial communications with the PC
  SPI.begin(); // Init SPI bus
  mfrc522.PCD_Init(); // Init MFRC522 card
  lcd.init(); // initialize the lcd
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("Starting...");
  lcd.setCursor(0, 1);
  lcd.print("Wait 3 beeps");
  pinMode(PASSIVE_BUZZER_PIN, OUTPUT); // Set passive buzzer pin as output
  pinMode(SERVO_PIN, OUTPUT); // Set servo pin as output
  myservo.attach(SERVO_PIN); // attaches the servo on pin to the servo object
  digitalWrite(SERVO_PIN, LOW); // Initially turn off servo
  delay(20000); // allow the MQ3 to warm up

  // Three beeps to indicate alcohol sensor heating
  for (int i = 0; i < 3; i++) {
    tone(PASSIVE_BUZZER_PIN, 1000); 
    delay(200); 
    noTone(PASSIVE_BUZZER_PIN); 
    delay(200);
  }
  
  lcd.clear();
}

void loop() {
  lcd.setCursor(0, 0);
  lcd.print("Use your card");
  lcd.setCursor(0, 1);
  lcd.print("to check ID");

  if (mfrc522.PICC_IsNewCardPresent() && mfrc522.PICC_ReadCardSerial()) {
    bool ID_known = false;
    String uid = getUID();
    uid.replace(" ", "");
    Serial.println("UID: " + uid);
     if (uid == "138bfd0d") { // Check if the UID matches the specific value
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Hello Marian");
      lcd.setCursor(0, 1);
      lcd.print("Age: 22 years");
      beepSuccess(); // Success beep
      ID_known = true;
    } else {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("ID NOT KNOWN");
      tone(PASSIVE_BUZZER_PIN, 1000); // Long beep for denial
      delay(1000); // Delay for beep duration
      noTone(PASSIVE_BUZZER_PIN); // Turn off the sound
    }
    delay(5000); 
    if (ID_known){
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Alcohol test");
    lcd.setCursor(0, 1);
    lcd.print("is starting...");

    delay(3000); // Delay to inform about the breathalyzer test

    float alcoholPercentage = performBreathalyzerTest();

    Serial.println("UID: " + uid);

    // Save data to SD card
    saveDataToSD(uid, alcoholPercentage);

    // Check if alcohol percentage is zero
    if (alcoholPercentage == 0) {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Alcohol free!");
      lcd.setCursor(0, 1);
      lcd.print("Starting engine...");
      delay(1000);
      beepSuccess();
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Engine started!");
      lcd.setCursor(0, 1);
      lcd.print("Ready to drive.");
      servoSpin();
      delay(1000000);
    }

    if (alcoholPercentage > 0) {
      beepFailed();
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Alcohol detected!");
      
      int remainingDelay = 120; // 2 minutes in seconds

      // Update LCD with remaining delay time
      while (remainingDelay > 0) {
        lcd.setCursor(0, 1);
        lcd.print("Delay: ");
        lcd.print(remainingDelay);
        lcd.print("s ");
        delay(1000); // Update every second
        remainingDelay--;
      }

      lcd.clear();
    }

    delay(5000); // Delay before re-enabling keycard scanning
    }
    else{
      lcd.clear();
    }
  }
}

float performBreathalyzerTest() {
  unsigned long startTime = millis(); 
  unsigned long testDuration = 20000; 
  unsigned long lastUpdateTime = startTime; 
  int remainingSeconds = (testDuration - (millis() - startTime)) / 1000;

  // Read baseline sensor value in clean air environment
  int baselineSensorValue = analogRead(MQ3pin);
  float alcoholConcentration_mgdl = 0;

  while (millis() - startTime < testDuration) {
    remainingSeconds = (testDuration - (millis() - startTime)) / 1000;

    // Read analog value from MQ3 sensor
    int sensorValue = analogRead(MQ3pin);
        Serial.println("sensor val: ");
    Serial.println(sensorValue);

    // Subtract baseline reading from current sensor value
    int correctedSensorValue = sensorValue - baselineSensorValue;

    if (correctedSensorValue > 100) {
      // Convert corrected sensor value to alcohol concentration in mg/dL using calibration factor
      alcoholConcentration_mgdl = 0.4 * correctedSensorValue;
      break;
    }

    // Update LCD every second
    if (millis() - lastUpdateTime >= 1000) {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Testing alcohol...");
      lcd.setCursor(0, 1);
      lcd.print("Time left: ");
      lcd.print(remainingSeconds);
      lcd.print("s");
      lastUpdateTime = millis(); // Update last update time
    }
           Serial.println("corrected val: ");
    Serial.println(correctedSensorValue);
    // Allow other tasks to execute by adding a short delay
    yield();
  }

  return alcoholConcentration_mgdl;
}


String getUID() {
  String uid = "";
  for (byte i = 0; i < mfrc522.uid.size; i++) {
    uid.concat(String(mfrc522.uid.uidByte[i] < 0x10 ? " 0" : " "));
    uid.concat(String(mfrc522.uid.uidByte[i], HEX));
  }
  return uid;
}

void saveDataToSD(String uid, float alcoholPercentage) {
  // Initialize SD card
  if (!SD.begin(chipSelect)) {
    Serial.println("SD initialization failed!");
    return;
  }

  // Open file for writing
  dataFile = SD.open("test.txt", FILE_WRITE);
  if (dataFile) {
    dataFile.print("UID: ");
    dataFile.println(uid);
    dataFile.print("Alcohol Percentage: ");
    dataFile.println(alcoholPercentage);
    dataFile.close();
    Serial.println("Data saved to SD card!");
  } else {
    Serial.println("Error opening data file!");
  }
}

void beepSuccess() {
  tone(PASSIVE_BUZZER_PIN, 1000); 
  delay(200); 
  noTone(PASSIVE_BUZZER_PIN); 
  delay(100);
}

void beepFailed() {
  for (int i = 0; i < 3; i++) {
    tone(PASSIVE_BUZZER_PIN, 500); 
    delay(200); 
    noTone(PASSIVE_BUZZER_PIN); 
    delay(200);
  }
}

void servoSpin() {
  for (int i = 0; i < 5; i++) {
    // Rotate from 180 degrees to 0 degrees
    for (int angle = 180; angle >= 0; angle--) {
      myservo.write(angle);
      delay(15); 
    }
    // Rotate back from 0 degrees to 180 degrees
    for (int angle = 0; angle <= 180; angle++) {
      myservo.write(angle);
      delay(15);
    }
  }
}
